@@ -25,6 +25,8 @@
 #include "pango-font.h"
 #include "pango-impl-utils.h"
 
+#include <hb-ot.h>
+
 /**
  * pango_glyph_string_new:
  *
@@ -368,8 +370,14 @@ pango_glyph_string_get_logical_widths (PangoGlyphString *glyphs,
  * Converts from character position to x position.
  *
  * The X position is measured from the left edge of the run.
- * Character positions are computed by dividing up each cluster
- * into equal portions.
+ * Character positions are obtained using font metrics for ligatures
+ * where available, and computed by dividing up each cluster
+ * into equal portions, otherwise.
+ *
+ * <picture>
+ *   <source srcset="glyphstring-positions-dark.png" media="(prefers-color-scheme: dark)">
+ *   <img alt="Glyph positions" src="glyphstring-positions-light.png">
+ * </picture>
  */
 void
 pango_glyph_string_index_to_x (PangoGlyphString *glyphs,
@@ -379,6 +387,45 @@ pango_glyph_string_index_to_x (PangoGlyphString *glyphs,
                                int               index,
                                gboolean          trailing,
                                int              *x_pos)
+{
+  pango_glyph_string_index_to_x_full (glyphs,
+                                      text, length,
+                                      analysis,
+                                      NULL,
+                                      index, trailing,
+                                      x_pos);
+}
+
+/**
+ * pango_glyph_string_index_to_x_full:
+ * @glyphs: the glyphs return from [func@shape]
+ * @text: the text for the run
+ * @length: the number of bytes (not characters) in @text.
+ * @analysis: the analysis information return from [func@itemize]
+ * @attrs: (nullable): `PangoLogAttr` array for @text
+ * @index_: the byte index within @text
+ * @trailing: whether we should compute the result for the beginning (%FALSE)
+ *   or end (%TRUE) of the character.
+ * @x_pos: (out): location to store result
+ *
+ * Converts from character position to x position.
+ *
+ * This variant of [method@Pango.GlyphString.index_to_x] additionally
+ * accepts a `PangoLogAttr` array. The grapheme boundary information
+ * in it can be used to disambiguate positioning inside some complex
+ * clusters.
+ *
+ * Since: 1.50
+ */
+void
+pango_glyph_string_index_to_x_full (PangoGlyphString *glyphs,
+                                    const char       *text,
+                                    int               length,
+                                    PangoAnalysis    *analysis,
+                                    PangoLogAttr     *attrs,
+                                    int               index,
+                                    gboolean          trailing,
+                                    int              *x_pos)
 {
   int i;
   int start_xpos = 0;
@@ -390,6 +437,8 @@ pango_glyph_string_index_to_x (PangoGlyphString *glyphs,
 
   int cluster_chars = 0;
   int cluster_offset = 0;
+  int start_glyph_pos = -1;
+  int end_glyph_pos = -1;
 
   const char *p;
 
@@ -406,51 +455,74 @@ pango_glyph_string_index_to_x (PangoGlyphString *glyphs,
       return;
     }
 
+  start_glyph_pos = -1;
+  end_glyph_pos = -1;
+
   /* Calculate the starting and ending character positions
    * and x positions for the cluster
    */
   if (analysis->level % 2) /* Right to left */
     {
       for (i = glyphs->num_glyphs - 1; i >= 0; i--)
-	width += glyphs->glyphs[i].geometry.width;
+        width += glyphs->glyphs[i].geometry.width;
 
       for (i = glyphs->num_glyphs - 1; i >= 0; i--)
-	{
-	  if (glyphs->log_clusters[i] > index)
-	    {
-	      end_index = glyphs->log_clusters[i];
-	      end_xpos = width;
-	      break;
-	    }
+        {
+          if (glyphs->log_clusters[i] > index)
+            {
+              end_index = glyphs->log_clusters[i];
+              end_xpos = width;
+              break;
+            }
 
-	  if (glyphs->log_clusters[i] != start_index)
-	    {
-	      start_index = glyphs->log_clusters[i];
-	      start_xpos = width;
-	    }
+          if (glyphs->log_clusters[i] != start_index)
+            {
+              start_index = glyphs->log_clusters[i];
+              start_xpos = width;
+            }
 
-	  width -= glyphs->glyphs[i].geometry.width;
-	}
+          width -= glyphs->glyphs[i].geometry.width;
+        }
+
+      for (i = glyphs->num_glyphs - 1; i >= 0; i--)
+        {
+          if (glyphs->log_clusters[i] == start_index)
+            {
+              if (end_glyph_pos < 0)
+                end_glyph_pos = i;
+              start_glyph_pos = i;
+            }
+        }
     }
   else /* Left to right */
     {
       for (i = 0; i < glyphs->num_glyphs; i++)
-	{
-	  if (glyphs->log_clusters[i] > index)
-	    {
-	      end_index = glyphs->log_clusters[i];
-	      end_xpos = width;
-	      break;
-	    }
+        {
+          if (glyphs->log_clusters[i] > index)
+            {
+              end_index = glyphs->log_clusters[i];
+              end_xpos = width;
+              break;
+            }
 
-	  if (glyphs->log_clusters[i] != start_index)
-	    {
-	      start_index = glyphs->log_clusters[i];
-	      start_xpos = width;
-	    }
+          if (glyphs->log_clusters[i] != start_index)
+            {
+              start_index = glyphs->log_clusters[i];
+              start_xpos = width;
+            }
 
-	  width += glyphs->glyphs[i].geometry.width;
-	}
+          width += glyphs->glyphs[i].geometry.width;
+        }
+
+      for (i = 0; i < glyphs->num_glyphs; i++)
+        {
+          if (glyphs->log_clusters[i] == start_index)
+            {
+              if (start_glyph_pos < 0)
+                start_glyph_pos = i;
+              end_glyph_pos = i;
+            }
+        }
     }
 
   if (end_index == -1)
@@ -459,28 +531,95 @@ pango_glyph_string_index_to_x (PangoGlyphString *glyphs,
       end_xpos = (analysis->level % 2) ? 0 : width;
     }
 
-  /* Calculate offset of character within cluster */
-
-  p = text + start_index;
-  while (p < text + end_index)
+  /* Calculate offset of character within cluster.
+   * To come up with accurate answers here, we need to know grapheme
+   * boundaries.
+   */
+  for (p = text + start_index, i = attrs ? g_utf8_pointer_to_offset (text, text + start_index) : 0;
+       p < text + end_index;
+       p = g_utf8_next_char (p), i++)
     {
+      if (attrs && !attrs[i].is_cursor_position)
+        continue;
+
       if (p < text + index)
-	cluster_offset++;
+        cluster_offset++;
       cluster_chars++;
-      p = g_utf8_next_char (p);
     }
 
   if (trailing)
     cluster_offset += 1;
 
   if (G_UNLIKELY (!cluster_chars)) /* pedantic */
-    {
+     {
       *x_pos = start_xpos;
       return;
     }
 
+  /* Try to get a ligature caret position for the glyph from the font.
+   * This only makes sense if the cluster contains a single spacing
+   * glyph, so we need to check that all but one of them are marks.
+   */
+  if (cluster_offset > 0 && cluster_offset < cluster_chars)
+    {
+      hb_font_t *hb_font;
+      hb_position_t caret;
+      unsigned int caret_count = 1;
+      int glyph_pos;
+      int num_carets;
+
+      hb_font = pango_font_get_hb_font (analysis->font);
+
+      if (start_glyph_pos == end_glyph_pos)
+        glyph_pos = start_glyph_pos;
+      else
+        {
+          hb_face_t *hb_face;
+
+          hb_face = hb_font_get_face (hb_font);
+
+          glyph_pos = -1;
+          for (i = start_glyph_pos; i <= end_glyph_pos; i++)
+            {
+              if (hb_ot_layout_get_glyph_class (hb_face, glyphs->glyphs[i].glyph) != HB_OT_LAYOUT_GLYPH_CLASS_MARK)
+                {
+                  if (glyph_pos != -1)
+                    {
+                      /* multiple non-mark glyphs in cluster, giving up */
+                      goto fallback;
+                    }
+                  glyph_pos = i;
+                }
+            }
+          if (glyph_pos == -1)
+            {
+              /* no non-mark glyph in a multi-glyph cluster, giving up */
+              goto fallback;
+            }
+        }
+
+      num_carets = hb_ot_layout_get_ligature_carets (hb_font,
+                                                     (analysis->level % 2) ? HB_DIRECTION_RTL : HB_DIRECTION_LTR,
+                                                      glyphs->glyphs[glyph_pos].glyph,
+                                                      cluster_offset - 1, &caret_count, &caret);
+      if (caret_count == 0 || num_carets == 0)
+        {
+          /* no ligature caret information found for this glyph */
+          goto fallback;
+        }
+
+      if (analysis->level % 2) /* Right to left */
+        *x_pos = end_xpos + caret;
+      else
+        *x_pos = start_xpos + caret;
+      *x_pos += glyphs->glyphs[glyph_pos].geometry.x_offset;
+      return;
+    }
+
+fallback:
+
   *x_pos = ((cluster_chars - cluster_offset) * start_xpos +
-	    cluster_offset * end_xpos) / cluster_chars;
+            cluster_offset * end_xpos) / cluster_chars;
 }
 
 /**
